<?php
/**
 * finc specific model for Solr records based on the stock
 * VuFind\RecordDriver\SolrDefault
 *
 * PHP version 5
 *
 * Copyright (C) Leipzig University Library 2015.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  RecordDrivers
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:record_drivers Wiki
 */

namespace finc\RecordDriver;

use VuFind\Db\Row\Resource;

/**
 * finc specific model for Missing records
 *
 * @category VuFind
 * @package  RecordDrivers
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:record_drivers Wiki
 *
 * @SuppressWarnings(PHPMD.ExcessivePublicCount)
 */
class FincMissing extends \VuFind\RecordDriver\Missing
{
    use FincRecordSerializationTrait;

    /**
     * @var array $availableTitleFields list of title fields in exploration order
     */
    protected $availableTitleFields = [
        'title',
        'title_full',
        'title_short',
        'title_sort'
    ];

    /**
     * @var string $displayTitle
     */
    protected $displayTitle;

    /**
     * @var string $shortTitle
     */
    protected $shortTitle;

    /**
     * Sets the raw data from the resource table. This is deserialized from the title field
     * @param mixed $data
     */
    public function setRawData($data)
    {
        if (isset($data['id']) && !empty($data['id'])) {
            $table = $this->getDbTable('resource');
            // sets create to FALSE so Resource does not try to generate a DB row on fail
            $resource = $table->findResource($data['id'], 'solr', FALSE);
            if ($resource instanceof Resource) {
                if (empty($resource->extra_metadata ?? '')) {
                    parent::setRawData($data);
                } else {
                    $this->deserializeFromResourceTable($resource);
                }
            } else {
                $this->fields['id'] = $data['id'];
            }
        } else {
            $this->fields['id'] = '';
        }
    }

    /**
     * each FincMissing Record is from cache
     * @return bool TRUE
     */
    public function isCachedRecord()
    {
        return TRUE;
    }

    /**
     * {@inheritDoc}
     * Overriden to enable the use of different title fields
     * @return mixed|string
     */
    public function determineMissingTitle()
    {
        if (!isset($this->displayTitle)) {
            foreach ($this->availableTitleFields as $field) {
                if (isset($this->fields[$field])) {
                    $this->displayTitle = $this->fields[$field];
                    break;
                }
            }
            if (!isset($this->displayTitle)) {
                $this->displayTitle = parent::determineMissingTitle();
            }
        }
        return $this->displayTitle;
    }

    /**
     * {@inheritDoc}
     * @return string
     */
    public function getRecordType()
    {
        return 'missing';
    }

    /**
     * {@inheritdoc}
     * adds some short term caching
     * @return mixed|string
     */
    public function getShortTitle()
    {
        if (!isset($this->shortTitle)) {
            if (isset($this->fields['title_short'])) {
                $this->shortTitle = $this->fields['title_short'];
            } else {
                $title = $this->determineMissingTitle();
                $this->shortTitle = trim(preg_replace('/[^\w\s]/u', '', $title));
            }
        }
        return $this->shortTitle;
    }

    /**
     * {@inheritoc}
     * @return array
     */
    public function getPlacesOfPublication()
    {
        if (isset($this->fields['publishPlace'])) {
            return (array)$this->fields['publishPlace'];
        }
        return [];
    }

    /**
     * {@inheritdoc}
     * Casts output to array
     * @return array
     */
    public function getPublicationDates()
    {
        return (array)parent::getPublicationDates();
    }

    /**
     * Returns the content of the additional_authors field
     * @return array|mixed
     */
    public function getAdditionalAuthors()
    {
        return $this->fields['additional_authors'] ?? [];
    }
}
